from bareasgi import Info, WebSocket
import graphql
import json
import logging
from typing import Any, Mapping, MutableMapping, Optional, AsyncIterator, Tuple, Union, List

logger = logging.getLogger(__name__)

WS_INTERNAL_ERROR = 1011
WS_PROTOCOL = "graphql-ws"

GQL_CONNECTION_INIT = "connection_init"  # Client -> Server
GQL_CONNECTION_ACK = "connection_ack"  # Server -> Client
GQL_CONNECTION_ERROR = "connection_error"  # Server -> Client
GQL_CONNECTION_KEEP_ALIVE = "ka"  # Server -> Client
GQL_CONNECTION_TERMINATE = "connection_terminate"  # Client -> Server
GQL_START = "start"  # Client -> Server
GQL_DATA = "data"  # Server -> Client
GQL_ERROR = "error"  # Server -> Client
GQL_COMPLETE = "complete"  # Server -> Client
GQL_STOP = "stop"  # Client -> Server


class ProtocolError(Exception):
    pass


class GraphQLWebSocketHandlerInstance:

    def __init__(self, schema: graphql.GraphQLSchema, web_socket: WebSocket, info: Info) -> None:
        self.schema = schema
        self.web_socket = web_socket
        self.info = info
        self._subscriptions: MutableMapping[int, AsyncIterator[graphql.ExecutionResult]] = {}
        self._is_closed = False

    async def start(self, subprotocols: List[str]):
        await self._on_open(subprotocols)

        _type = GQL_CONNECTION_KEEP_ALIVE
        try:
            while _type not in (GQL_CONNECTION_ERROR, GQL_CONNECTION_TERMINATE):
                type_, id_, payload = await self._read_message()
                await self._on_message(type_, id_, payload)

        except ProtocolError as error:
            logger.error(f'Failed to process message: {error}')
        except EOFError:
            self._is_closed = True
        except Exception as error:
            logger.error(f'Internal error: {error}')

        await self._on_close()

    async def _on_open(self, subprotocols: List[str]) -> None:
        if not WS_PROTOCOL in subprotocols:
            raise ProtocolError(f"Expected subprotocol '{WS_PROTOCOL}")
        await self.web_socket.accept(WS_PROTOCOL)

    async def _on_close(self) -> None:
        await self._unsubscribe_all()
        if not self._is_closed:
            await self.web_socket.close()

    async def _read_message(self) -> Tuple[str, Optional[int], Optional[dict]]:
        text = await self.web_socket.receive()
        if text is None:
            raise EOFError

        if not isinstance(text, str):
            raise ProtocolError('Expected the message to be a string.')

        message = json.loads(text)
        if not isinstance(message, dict):
            raise ProtocolError('Expected the message to be an object.')

        type_ = message['type']
        if not isinstance(type_, str):
            raise ProtocolError("Expected field 'type' to be a string")

        id_ = message.get('id')
        if not (id_ is None or isinstance(id_, int)):
            raise ProtocolError("Expected field 'id' to be an int?")

        payload = message.get('payload')
        if not (payload is None or isinstance(payload, dict)):
            raise ProtocolError("Expected field 'payload' to be an object?")

        return type_, id_, payload

    async def _on_message(self, type_: str, id_: Optional[int], payload: Optional[dict]):

        if type_ == GQL_CONNECTION_INIT:
            await self._on_connection_init(id_, payload)
        elif type_ == GQL_CONNECTION_TERMINATE:
            await self._on_connection_terminate()
        elif type_ == GQL_START:
            await self._on_start(id_, payload)
        elif type_ == GQL_STOP:
            await self._on_stop(id_)
        elif type_ == GQL_CONNECTION_KEEP_ALIVE:
            pass
        else:
            raise ProtocolError(f"Received unknown message type '{type_}'.")

    async def _on_connection_init(self, id_: Optional[int], connection_params: Optional[Any]):
        try:
            await self.web_socket.send(self.to_message('connection_ack', id_))
        except Exception as error:
            await self._send_error(GQL_CONNECTION_ERROR, id_, error)
            await self.web_socket.close(WS_INTERNAL_ERROR)
            raise

    async def _on_connection_terminate(self):
        await self.web_socket.close(WS_INTERNAL_ERROR)

    async def _on_start(self, id_: Optional[int], payload: Optional[Union[list, dict]]):
        try:
            # An id is required for a start operation.
            if id_ is None:
                raise ProtocolError("required 'id' field must be an int.")

            if id_ in self._subscriptions:
                await self._unsubscribe(id_)
                del self._subscriptions[id_]

            query, variable_values, operation_name = self.parse_start_payload(payload)

            document = graphql.parse(query)
            if any(definition.operation is graphql.OperationType.SUBSCRIPTION for definition in document.definitions):
                result = await graphql.subscribe(
                    self.schema,
                    document=document,
                    context_value=self.info,
                    variable_values=variable_values,
                    operation_name=operation_name
                )
            else:
                result = await graphql.graphql(
                    self.schema,
                    source=graphql.Source(query),
                    context_value=self.info,
                    variable_values=variable_values,
                    operation_name=operation_name
                )

            if not isinstance(result, AsyncIterator):
                await self._send_execution_result(id_, result)
                return True

            self._subscriptions[id_] = result
            try:
                async for val in result:
                    await self._send_execution_result(id_, val)
            finally:
                del self._subscriptions[id_]
                await self.web_socket.send(self.to_message(GQL_COMPLETE, id_))

        except Exception as error:
            await self._send_error(GQL_ERROR, id_, error)

    async def _on_stop(self, id_: int):
        await self._unsubscribe(id_)

    async def _unsubscribe(self, id_: int) -> None:
        await self._subscriptions[id_].aclose()

    async def _unsubscribe_all(self):
        for id_ in self._subscriptions.keys():
            await self._unsubscribe(id_)

    async def _send_error(self, type_: str, id_: Optional[int], error: Exception) -> None:
        await self.web_socket.send(self.to_message(type_, id_, {'message': str(error)}))

    async def _send_execution_result(self, id_: int, execution_result: graphql.ExecutionResult) -> None:
        result = dict()

        if execution_result.data:
            result["data"] = execution_result.data

        if execution_result.errors:
            result["errors"] = [graphql.format_error(error) for error in execution_result.errors]

        await self.web_socket.send(self.to_message(GQL_DATA, id_, result))

    @classmethod
    def to_message(cls, type_: str, id_: Optional[int] = None, payload: Optional[Any] = None) -> str:
        message = {'type': type_}
        if id_ is not None:
            message['id'] = id_
        if payload is not None:
            message['payload'] = payload
        return json.dumps(message)

    @classmethod
    def parse_start_payload(
            cls,
            payload: Optional[Union[dict, list]]
    ) -> Tuple[str, Optional[Mapping[str, Any]], Optional[str]]:

        if not isinstance(payload, dict):
            raise ProtocolError("required 'payload' field must be an object.")

        query = payload.get('query')
        if not isinstance(query, str):
            raise ProtocolError("required 'query' field must be string in 'payload'.")

        variable_values = payload.get('variables')
        if not (variable_values is None or isinstance(variable_values, dict)):
            raise ProtocolError("optional 'variables' field must be object? in 'payload'.")

        operation_name = payload.get('operationName')
        if not (operation_name is None or isinstance(operation_name, str)):
            raise ProtocolError("optional 'operationName' field must be str? in 'payload'.")

        return query, variable_values, operation_name



from cgi import parse_multipart
# noinspection PyPackageRequirements
from graphql import graphql, GraphQLSchema
import io
from typing import List
from urllib.parse import parse_qs
import json
from bareasgi import (
    Application,
    Scope,
    Info,
    RouteMatches,
    Content,
    HttpResponse,
    WebSocket,
    text_writer,
    text_reader
)
from bareasgi.types import Header
from bareasgi.middleware import mw

from .template import make_template
from .websocket_handler import GraphQLWebSocketHandler
from .utils import parse_header


class GraphQLController:

    def __init__(self, schema: GraphQLSchema, path_prefix: str = '', middleware=None):
        self.schema = schema
        self.path_prefix = path_prefix
        self.middleware = middleware
        self.subscription_handler = GraphQLWebSocketHandler(schema)

    # noinspection PyUnusedLocal
    async def view_graphiql(self, scope: Scope, info: Info, matches: RouteMatches, content: Content) -> HttpResponse:
        host, port = scope['server']
        body = make_template(f'{host}:{port}', self.path_prefix + '/graphql', self.path_prefix + '/subscriptions')
        headers = [
            (b'content-type', b'text/html'),
            (b'content-length', str(len(body)).encode())
        ]
        return 200, headers, text_writer(body)

    async def handle_subscription(self, scope: Scope, info: Info, matches: RouteMatches, web_socket: WebSocket) -> None:
        await self.subscription_handler(scope, info, matches, web_socket)

    @classmethod
    async def _get_query_document(cls, headers: List[Header], content: Content):
        content_type = next((v for k, v in headers if k == b'content-type'), None)
        content_type, parameters = parse_header(content_type)

        if content_type == b'application/graphql':
            return {'query': await text_reader(content)}
        elif content_type in (b'application/json', b'text/plain'):
            return json.loads(await text_reader(content))
        elif content_type == b'application/x-www-form-urlencoded':
            return parse_qs(await text_reader(content))
        elif content_type == b'multipart/form-data':
            return parse_multipart(io.StringIO(await text_reader(content)), parameters["boundary"])
        else:
            raise RuntimeError('Content type not supported')

    # noinspection PyUnusedLocal
    async def handle_graphql(self, scope: Scope, info: Info, matches: RouteMatches, content: Content) -> HttpResponse:

        # noinspection PyBroadException
        try:
            query_document = await self._get_query_document(scope['headers'], content)

            result = await graphql(
                schema=self.schema,
                source=query_document['query'],
                variable_values=query_document.get('variables'),
                operation_name=query_document.get('operationName'),
                context_value=info,
                middleware=self.middleware
            )

            response = {'data': result.data}
            if result.errors:
                response['errors'] = [error.formatted for error in result.errors]

            text = json.dumps(response)
            headers = [
                (b'content-type', b'application/json'),
                (b'content-length', str(len(text)).encode())
            ]

            return 200, headers, text_writer(text)

        except:
            text = 'Internal server error'
            headers = [
                (b'content-type', b'text/plain'),
                (b'content-length', str(len(text)).encode())
            ]
            return 500, headers, text_writer(text)


def add_graphql_next(
        app: Application,
        schema: GraphQLSchema,
        path_prefix: str = '',
        rest_middleware=None,
        view_middleware=None,
        graphql_middleware=None
) -> None:
    controller = GraphQLController(schema, path_prefix, graphql_middleware)

    # Add the REST route
    app.http_router.add(
        {'GET'},
        path_prefix + '/graphql',
        controller.handle_graphql if rest_middleware is None else mw(rest_middleware, handler=controller.handle_graphql)
    )
    app.http_router.add(
        {'POST', 'OPTION'},
        path_prefix + '/graphql',
        controller.handle_graphql if rest_middleware is None else mw(rest_middleware, handler=controller.handle_graphql)
    )

    # Add the subscription route
    app.ws_router.add(
        path_prefix + '/subscriptions',
        controller.handle_subscription
    )

    # Add Graphiql
    app.http_router.add(
        {'GET'},
        path_prefix + '/graphiql',
        controller.view_graphiql if view_middleware is None else mw(view_middleware, handler=controller.view_graphiql)
    )



import string

graphiql_template = string.Template(
    """
<!DOCTYPE html>
<html>
  <head>
    <style>
      body {
        height: 100%;
        margin: 0;
        width: 100%;
        overflow: hidden;
      }
      #graphiql {
        height: 100vh;
      }
    </style>
    
    <!--
      This GraphiQL example depends on Promise and fetch, which are available in
      modern browsers, but can be "polyfilled" for older browsers.
      GraphiQL itself depends on React DOM.
      If you do not want to rely on a CDN, you can host these files locally or
      include them directly in your favored resource bunder.
    -->
    <script src="//cdn.jsdelivr.net/es6-promise/4.0.5/es6-promise.auto.min.js"></script>
    <script src="//cdn.jsdelivr.net/fetch/0.9.0/fetch.min.js"></script>
    <script src="//cdn.jsdelivr.net/react/15.4.2/react.min.js"></script>
    <script src="//cdn.jsdelivr.net/react/15.4.2/react-dom.min.js"></script>
    
    <script src="//unpkg.com/subscriptions-transport-ws@${subscriptions_transport_version}/browser/client.js"></script>
    <script src="//unpkg.com/graphiql-subscriptions-fetcher@0.0.2/browser/client.js"></script>

    <link rel="stylesheet" href="https://cdn.jsdelivr.net/npm/graphiql-custom-headers@1.0.0/graphiql-custom-headers.css" />
    <script src="//cdn.jsdelivr.net/npm/graphiql-custom-headers@1.0.0/graphiql-custom-headers.min.js"></script>

  </head>
  <body>
    <div id="graphiql">Loading...</div>
    <script>
      /**
       * This GraphiQL example illustrates how to use some of GraphiQL's props
       * in order to enable reading and updating the URL parameters, making
       * link sharing of queries a little bit easier.
       *
       * This is only one example of this kind of feature, GraphiQL exposes
       * various React params to enable interesting integrations.
       */
       
      // Parse the search string to get url parameters.
      var search = window.location.search;
      var parameters = {};
      search.substr(1).split('&').forEach(function (entry) {
        var eq = entry.indexOf('=');
        if (eq >= 0) {
          parameters[decodeURIComponent(entry.slice(0, eq))] =
            decodeURIComponent(entry.slice(eq + 1));
        }
      });
      // if variables was provided, try to format it.
      if (parameters.variables) {
        try {
          parameters.variables =
            JSON.stringify(JSON.parse(parameters.variables), null, 2);
        } catch (e) {
          // Do nothing, we want to display the invalid JSON as a string, rather
          // than present an error.
        }
      }
      // When the query and variables string is edited, update the URL bar so
      // that it can be easily shared
      function onEditQuery(newQuery) {
        parameters.query = newQuery;
        updateURL();
      }
      function onEditVariables(newVariables) {
        parameters.variables = newVariables;
        updateURL();
      }
      function onEditOperationName(newOperationName) {
        parameters.operationName = newOperationName;
        updateURL();
      }
      function updateURL() {
        var newSearch = '?' + Object.keys(parameters).filter(function (key) {
          return Boolean(parameters[key]);
        }).map(function (key) {
          return encodeURIComponent(key) + '=' +
            encodeURIComponent(parameters[key]);
        }).join('&');
        history.replaceState(null, null, newSearch);
      }
      
      // Defines a GraphQL fetcher using the fetch API. You're not required to
      // use fetch, and could instead implement graphQLFetcher however you like,
      // as long as it returns a Promise or Observable.
      function graphQLFetcher(graphQLParams, customHeaders) {
        // This example expects a GraphQL server at the path /graphql.
        // Change this to point wherever you host your GraphQL server.
        return fetch(`${query_path}`, {
          method: 'post',
          headers: customHeaders,
          body: JSON.stringify(graphQLParams),
          credentials: 'include',
        }).then(function (response) {
          return response.text();
        }).then(function (responseBody) {
          try {
            return JSON.parse(responseBody);
          } catch (error) {
            return responseBody;
          }
        });
      }
      
      const websocketScheme = window.location.protocol === 'http:' ? 'ws:' : 'wss:'
      let subscriptionsClient = new window.SubscriptionsTransportWs.SubscriptionClient(
        `$${websocketScheme}//${host}${subscription_path}`, { reconnect: true });
      let subscriptionsFetcher = window.GraphiQLSubscriptionsFetcher.graphQLFetcher(subscriptionsClient, graphQLFetcher);
      
      // Render <GraphiQL /> into the body.
      // See the README in the top level of this module to learn more about
      // how you can customize GraphiQL by providing different values or
      // additional child elements.
      ReactDOM.render(
        React.createElement(GraphiQL, {
          fetcher: subscriptionsFetcher,
          query: parameters.query,
          variables: parameters.variables,
          operationName: parameters.operationName,
          onEditQuery: onEditQuery,
          onEditVariables: onEditVariables,
          onEditOperationName: onEditOperationName
        }),
        document.getElementById('graphiql')
      );
    </script>
  </body>
</html>
"""
)


def make_template(
        host: str,
        query_path: str = '/graphql',
        subscription_path: str = '/subscriptions',
        subscriptions_transport_version: str = "0.7.0",
) -> str:
    return graphiql_template.substitute(
        host=host,
        query_path=query_path,
        subscription_path=subscription_path,
        subscriptions_transport_version=subscriptions_transport_version,
    )



from typing import Tuple, Mapping, MutableMapping, Iterator


def _parseparam(s: bytes) -> Iterator[bytes]:
    while s[:1] == b';':
        s = s[1:]
        end = s.find(b';')
        while end > 0 and (s.count(b'"', 0, end) - s.count(b'\\"', 0, end)) % 2:
            end = s.find(b';', end + 1)
        if end < 0:
            end = len(s)
        f = s[:end]
        yield f.strip()
        s = s[end:]


def parse_header(line: bytes) -> Tuple[bytes, Mapping[bytes, bytes]]:
    """Parse a Content-type like header.

    Return the main content-type and a dictionary of options.

    """
    parts = _parseparam(b';' + line)
    key = parts.__next__()
    params: MutableMapping[bytes, bytes] = {}
    for p in parts:
        i = p.find(b'=')
        if i >= 0:
            name = p[:i].strip().lower()
            value = p[i + 1:].strip()
            if len(value) >= 2 and value[0] == value[-1] == '"':
                value = value[1:-1]
                value = value.replace(b'\\\\', b'\\').replace(b'\\"', b'"')
            params[name] = value
    return key, params



# Always prefer setuptools over distutils
from setuptools import setup, find_packages
# To use a consistent encoding
from codecs import open
from os import path

here = path.abspath(path.dirname(__file__))

# Get the long description from the README file
with open(path.join(here, 'README.md'), encoding='utf-8') as f:
    long_description = f.read()

# Load the requirements from the text file.
with open(path.join(here, 'requirements.txt'), encoding='utf-8') as f:
    requirements = [x.strip() for x in f.readlines() if x.strip()]

__version__ = "1.1.0"

setup(
    name='bareasgi_graphql_next',
    version=__version__,
    description='GraphQL support for the bareASGI framework',
    long_description=long_description,
    long_description_content_type='text/markdown',
    url='https://github.com/rob-blackbourn/bareasgi_graphql_next',
    author='Rob Blackbourn',
    author_email='rob.blackbourn@googlemail.com',
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3 :: Only',
        'Topic :: Software Development',
        'Topic :: Software Development :: Libraries'
    ],
    license='Apache 2',
    keywords='ASGI framework GraphQL',
    package_dir={'': 'src'},
    packages=find_packages(where='src', exclude=['tests', 'examples']),
    setup_requires=['pytest-runner'],
    install_requires=requirements,
    tests_require=['pytest']
)



from bareasgi import (
    Scope,
    Info,
    RouteMatches,
    WebSocket
)
import graphql

from .websocket_instance import GraphQLWebSocketHandlerInstance


class GraphQLWebSocketHandler:

    def __init__(self, schema: graphql.GraphQLSchema):
        self.schema = schema

    async def __call__(self, scope: Scope, info: Info, matches: RouteMatches, web_socket: WebSocket) -> None:
        instance = GraphQLWebSocketHandlerInstance(self.schema, web_socket, info)
        await instance.start(scope['subprotocols'])
